import logging
import os
import time

import requests
from monotonic import monotonic
from requests.packages.urllib3.util.retry import (  # pylint: disable=import-error
    Retry,
)
from six.moves.urllib.parse import urljoin

LOG = logging.getLogger("pubtools-exodus")
LOG_FORMAT = "%(asctime)s [%(levelname)-8s] %(message)s"


class ExodusGatewaySession(
    object
):  # pylint: disable=too-many-instance-attributes
    """Base class for operations passing through exodus-gateway."""

    def __init__(self, exodus_enabled=None):
        super(ExodusGatewaySession, self).__init__()

        self.gw_env = None
        self.gw_url = None
        self.gw_crt = None
        self.gw_key = None

        self.session = None
        self.publish = None

        self._exodus_enabled = exodus_enabled

        # These defaults are not advertised or expected but can be controlled
        # by environment variables when needed (e.g., testing).
        self.retries = int(os.getenv("EXODUS_GW_RETRIES") or "5")
        self.timeout = int(os.getenv("EXODUS_GW_TIMEOUT") or "900")
        self.wait = int(os.getenv("EXODUS_GW_WAIT") or "5")

    def new_session(self):
        retry_strategy = Retry(
            total=int(self.retries),
            backoff_factor=1,
            status_forcelist=[429, 500, 502, 503, 504],
        )
        adapter = requests.adapters.HTTPAdapter(max_retries=retry_strategy)

        out = requests.Session()
        out.cert = (self.gw_crt, self.gw_key)
        out.mount(self.gw_url, adapter)

        return out

    def unpack_response(self, response):
        """Raise if response was not successful.

        This is the same as response.raise_for_status(), merely wrapping it
        to ensure the body is logged when possible.
        """

        try:
            response.raise_for_status()
        except Exception as outer:
            try:
                body = response.json()
            except:
                raise outer

            LOG.error(
                "Unsuccessful response from exodus-gw: %s",
                body.get("detail") or body,
            )
            raise

    def do_request(self, **kwargs):
        if not self.session:
            self.session = self.new_session()

        resp = self.session.request(**kwargs)
        self.unpack_response(resp)
        return resp

    def check_cert(self):
        """Issue request to exodus-gw to identify permissions."""

        auth_url = urljoin(self.gw_url, "/whoami")
        resp = self.do_request(method="GET", url=auth_url)
        context = resp.json()

        for user_type, ident in (
            ("client", "serviceAccountId"),
            ("user", "internalUsername"),
        ):
            typed_ctx = context[user_type]
            if typed_ctx["authenticated"]:
                roles = [str(role) for role in typed_ctx["roles"]]
                LOG.debug(
                    "Authenticated with exodus-gw at %s as %s %s (roles: %s)",
                    self.gw_url,
                    user_type,
                    typed_ctx[ident],
                    roles,
                )
                break
        else:
            LOG.debug("Not authenticated with exodus-gw at %s", self.gw_url)

    def new_publish(self):
        """Issue request to exodus-gw to create a new publish."""

        if not self.exodus_enabled:
            return None

        self._populate_exodus_gw_vars()
        self.check_cert()

        publish_url = os.path.join(self.gw_url, self.gw_env, "publish")
        resp_json = self.do_request(method="POST", url=publish_url).json()

        LOG.info("Created exodus-gw publish %s", resp_json["id"])

        return resp_json

    def poll_commit_completion(self, commit):
        """Issues request(s) to exodus-gw for the commit's state, returning
        if/when the state is either "COMPLETE" or "FAILED".
        """

        timelimit = monotonic() + self.timeout

        msg = "exodus-gw commit %s to %s" % (commit["id"], self.gw_url)

        while monotonic() < timelimit:
            task_url = urljoin(self.gw_url, commit["links"]["self"])
            resp = self.do_request(method="GET", url=task_url)
            task = resp.json()

            if task["state"] == "COMPLETE":
                LOG.debug("%s complete", msg)
                return task
            if task["state"] == "FAILED":
                raise RuntimeError("%s failed" % msg)

            time.sleep(self.wait)

        raise RuntimeError("Polling for %s timed out" % msg)

    def commit_publish(self, publish):
        """Commits an exodus-gw publish, e.g.,
        https://exodus-gw.example.com/prod/publish/4e59c1a0/commit
        """

        LOG.info("Committing exodus-gw publish %s", publish["id"])

        commit_url = urljoin(self.gw_url, publish["links"]["commit"])
        resp = self.do_request(method="POST", url=commit_url)
        commit = resp.json()

        self.poll_commit_completion(commit)

        LOG.info("Committed exodus-gw publish %s", publish["id"])

    @property
    def exodus_enabled(self):
        if self._exodus_enabled is None:
            enable_vals = ["true", "t", "1", "yes", "y"]
            self._exodus_enabled = (
                os.getenv("EXODUS_ENABLED", "False").lower() in enable_vals
            )
        return self._exodus_enabled

    def _populate_exodus_gw_vars(self):
        """Populate exodus gateway details from environment variables. All exodus CDN transactions
        go through exodus gateway."""

        self.gw_env = os.getenv("EXODUS_GW_ENV")
        if not self.gw_env:
            raise RuntimeError(
                "Environment variable '%s' is not set" % "EXODUS_GW_ENV"
            )

        self.gw_url = os.getenv("EXODUS_GW_URL")
        if not self.gw_url:
            raise RuntimeError(
                "Environment variable '%s' is not set" % "EXODUS_GW_URL"
            )

        self.gw_crt = os.getenv("EXODUS_GW_CERT")
        if not self.gw_crt:
            raise RuntimeError(
                "Environment variable '%s' is not set" % "EXODUS_GW_CERT"
            )

        self.gw_key = os.getenv("EXODUS_GW_KEY")
        if not self.gw_key:
            raise RuntimeError(
                "Environment variable '%s' is not set" % "EXODUS_GW_KEY"
            )



import logging
import os
import subprocess

import attr
from pushsource import Source

from pubtools.exodus.task import ExodusTask

LOG = logging.getLogger("pubtools-exodus")
LOG_FORMAT = "%(asctime)s [%(levelname)-8s] %(message)s"


class ExodusPushTask(ExodusTask):
    """Push a directory to the Exodus CDN"""

    def add_args(self):
        super(ExodusPushTask, self).add_args()

        self.parser.add_argument(
            "source",
            help=(
                """Source(s) of content to be pushed (e.g., 'staged:/path/to/staging/root')."""
            ),
        )

    @property
    def push_items(self):
        with Source.get(self.args.source) as source:
            for item in source:
                if item.src and len(item.dest) == 1:
                    # If the source directory passed to rsync does not end in a trailing
                    # slash ("/"), rsync will publish the entire directory, not just the
                    # contents of the directory.
                    #
                    # For example, "rsync ./example/rhel-8-for-x86_64-baseos-kickstart__7_DOT_4/RAW
                    # exodus:rhel-8-for-x86_64-baseos-kickstart__8_DOT_4" would
                    # publish to "/content/dist/rhel8/8.4/x86_64/baseos/kickstart/RAW",
                    # whereas "rsync ./example/rhel-8-for-x86_64-baseos-kickstart__8_DOT_4/RAW/
                    # exodus:rhel-8-for-x86_64-baseos-kickstart__8_DOT_4" would publish to
                    # "/content/dist/rhel8/8.4/x86_64/baseos/kickstart".
                    #
                    # See https://linux.die.net/man/1/rsync for more details.
                    item_src = item.src
                    if os.path.isdir(item.src) and not item.src.endswith("/"):
                        item_src = "%s/" % item.src
                    yield attr.evolve(item, src=item_src)
                else:
                    LOG.warning("Unexpected push item type: %s", item)

    def run(self):
        LOG.debug("Exodus push begins")

        publish = self.new_publish()
        publish_id = str(publish.get("id"))
        LOG.info("Publish ID: %s", publish_id)

        for item in self.push_items:
            LOG.debug("Processing %s", item)
            cmd = [
                "exodus-rsync",
                "--exodus-publish",
                publish_id,
                "--exclude",
                ".nfs*",
                "--exclude",
                ".latest_rsync",
                "--exclude",
                ".lock",
                item.src,
                "exodus:%s" % item.dest[0],
            ]
            if self.args.verbose:
                cmd.append("-" + "v" * self.args.verbose)
            if self.extra_args:
                cmd.extend(self.extra_args)

            LOG.info(" ".join(cmd))

            proc = subprocess.Popen(
                cmd,
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                universal_newlines=True,
            )

            for line in iter(proc.stdout.readline, ""):
                LOG.info(line.strip())

            ret = proc.wait()
            if ret != 0:
                raise RuntimeError("Exodus push failed")

        self.commit_publish(publish)

        LOG.info("Exodus push is complete")


def entry_point(args=None):
    task = ExodusPushTask(args)
    task.main()


def doc_parser():
    return ExodusPushTask().parser



import logging
import sys
from threading import Lock

import attr
from pubtools.pluggy import hookimpl, pm  # pylint: disable=wrong-import-order

from ..gateway import ExodusGatewaySession

LOG = logging.getLogger("pubtools-exodus")

# pylint: disable=unused-argument


class ExodusPulpHandler(ExodusGatewaySession):
    def __init__(self):
        super(ExodusPulpHandler, self).__init__()

        self.lock = Lock()
        self.publish_committed = False

    @hookimpl
    def pulp_repository_pre_publish(self, repository, options):
        """Invoked as the first step in publishing a Pulp repository.

        This implementation adds to each config the --exodus-publish argument,
        attaching the repository to an exodus-gw publish.

        Args:
            repository (:class:`~pubtools.pulplib.Repository`):
                The repository to be published.
            options (:class:`~pubtools.pulplib.PublishOptions`):
                The options to use in publishing.
        Returns:
            options (:class:`~pubtools.pulplib.PublishOptions`):
                The adjusted options used for this publish.
        """

        with self.lock:
            if not self.publish:
                self.publish = self.new_publish()

        if not self.publish:
            return None

        args = (
            list(options.rsync_extra_args) if options.rsync_extra_args else []
        )
        args.append("--exodus-publish=%s" % self.publish["id"])
        return attr.evolve(options, rsync_extra_args=args)

    def ensure_publish_committed(self):
        """Commit the current publish, if and only if there is a current publish
        and it was not already committed.
        """
        if not self.publish:
            LOG.debug("No exodus-gw publish to commit")
            return

        if self.publish_committed:
            LOG.debug("Publish is already committed")
            return

        self.publish_committed = True
        self.commit_publish(self.publish)

    @hookimpl
    def task_pulp_flush(self):
        """Invoked during task execution after successful completion of all
        Pulp publishes.

        This implementation commits the active exodus-gw publish, making
        the content visible on the target CDN environment.
        """
        self.ensure_publish_committed()

    @hookimpl
    def task_stop(self, failed):
        pm.unregister(self)
        if not failed:
            # If a task is finishing up successfully and there's still an uncommitted
            # publish, make sure to commit it. This is not expected to do anything
            # if the commit happened earlier via task_pulp_flush.
            self.ensure_publish_committed()


@hookimpl
def task_start():
    pm.register(ExodusPulpHandler())


pm.register(sys.modules[__name__])



from setuptools import find_packages, setup


def get_description():
    return "Utilities for interfacing with components of Red Hat's Content Delivery Network"


def get_long_description():
    with open("README.md") as readme:
        text = readme.read()

    # Long description is everything after README's initial heading
    idx = text.find("\n\n")
    return text[idx:]


def get_requirements():
    with open("requirements.txt") as reqs:
        return reqs.read().splitlines()


setup(
    name="pubtools-exodus",
    version="1.3.2",
    packages=find_packages(exclude=["tests"]),
    include_package_data=True,
    url="https://github.com/release-engineering/pubtools-exodus",
    license="GNU General Public License",
    description=get_description(),
    long_description=get_long_description(),
    long_description_content_type="text/markdown",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.6",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    install_requires=get_requirements(),
    python_requires=">=2.6",
    entry_points={
        "pubtools.hooks": [
            "pubtools-exodus-pulp = pubtools.exodus._hooks.pulp",
        ],
        "console_scripts": [
            "pubtools-exodus-push = pubtools.exodus._tasks.push:entry_point"
        ],
    },
    project_urls={
        "Changelog": "https://github.com/release-engineering/pubtools-exodus/blob/main/CHANGELOG.md",
        "Documentation": "https://release-engineering.github.io/pubtools-exodus",
    },
)



import logging
from argparse import ArgumentParser, RawDescriptionHelpFormatter

from pubtools.pluggy import task_context

from .gateway import ExodusGatewaySession

LOG = logging.getLogger("pubtools-exodus")
LOG_FORMAT = "%(asctime)s [%(levelname)-8s] %(message)s"


class ExodusTask(ExodusGatewaySession):
    """Base class for Exodus tasks"""

    def __init__(self, args=None):
        super(ExodusTask, self).__init__(exodus_enabled=True)

        self._args = None
        self._extra_args = None
        self._override_args = args

        self.parser = ArgumentParser(
            formatter_class=RawDescriptionHelpFormatter
        )
        self._basic_args()
        self.add_args()

    @property
    def args(self):
        """Parsed args from the cli
        Return the args if available from previous parse
        else parse with defined options and return the args
        """
        if self._args is None:
            # Parse a copy of the raw args, as the parse may modify them.
            to_parse = (
                list(self._override_args) if self._override_args else None
            )
            # If args are unset, it's likely extra_args are too.
            # Set both here to avoid repeating the operation later.
            self._args, self._extra_args = self.parser.parse_known_args(
                to_parse
            )
        return self._args

    @property
    def extra_args(self):
        """Remaining args from the cli
        Return the remaining args from the previous parse,
        else parse and return the remaining args
        """
        if self._extra_args is None:
            # Parse a copy of the raw args, as the parse may modify them.
            to_parse = (
                list(self._override_args) if self._override_args else None
            )
            _, self._extra_args = self.parser.parse_known_args(to_parse)
        return self._extra_args

    def _basic_args(self):
        # minimum args required for a CLI task
        self.parser.add_argument(
            "--debug",
            "-d",
            action="store_true",
            help=("Show debug logs"),
        )
        self.parser.add_argument(
            "--verbose",
            "-v",
            action="count",
            default=0,
            help=("Increase verbosity"),
        )

    def add_args(self):
        """Add parser options/arguments for a task
        e.g. self.parser.add_argument("option", help="help text")
        """
        # Calling super add_args if it exists allows this class and
        # Service classes to be inherited in either order without breaking.
        from_super = getattr(super(ExodusTask, self), "add_args", lambda: None)
        from_super()

    def _setup_logging(self):
        logging.basicConfig(level=logging.INFO, format=LOG_FORMAT)
        if self.args.debug:
            logging.getLogger("pubtools-exodus").setLevel(logging.DEBUG)

    def run(self):
        """Implement a specific task"""

        raise NotImplementedError()

    def main(self):
        """Main method called by the entrypoint of the task."""

        with task_context():
            # setup the logging as required
            self._setup_logging()

            self.run()
            return 0
